module LiteralExpression (LiteralExpression(..))
where

import Data.Ratio ((%))
import Data.Int (Int16)
import qualified Data.Char as Char

import JavaLanguage (m_lexeme, m_reserved)

import Text.Parsec (many1, digit, oneOf, option, char, choice, many, try, satisfy, manyTill, string', endOfLine, space)
import Text.Parsec.String (Parser)

data LiteralExpression = NullLiteral
        | IntegerLiteral   Integer  IntegerLiteralType
        | FloatingLiteral  Rational DecimalLiteralType
        | BooleanLiteral   Bool
        | CharacterLiteral Char
        | StringLiteral    String
        | TextLiteral      String
        deriving Show

data IntegerLiteralType = IntLiteral    | LongLiteral
        deriving Show
data DecimalLiteralType = DoubleLiteral | FloatLiteral
        deriving Show

signedInteger :: Parser Integer
signedInteger = do
        signFactor <- option 1 (char '-' *> return (-1))
        digits     <- many1 digit
        return (signFactor * read digits)

parseNullLiteral :: Parser LiteralExpression
parseNullLiteral = m_reserved "null" *> return NullLiteral

-- TODO: Hex, Oct, Bin
parseDecimalIntegerLiteral :: Parser LiteralExpression
parseDecimalIntegerLiteral = m_lexeme $ do
        negative    <- option False (char '-' *> return True)
        digitString <- many1 digit
        literalType <- option IntLiteral (oneOf "Ll" *> return LongLiteral)
        let positiveNumber = read digitString :: Integer
        let number = (if (negative) then negate else id) positiveNumber
        return (IntegerLiteral number literalType)

-- TODO: Hex
parseDecimalFloatingLiteral :: Parser LiteralExpression
parseDecimalFloatingLiteral = do
        prefix      <- option "0" $ many1 digit
        _           <- option '.' (char '.')
        suffix      <- many digit
        exponent    <- option 0 (char 'e' *> signedInteger)
        literalType <- option FloatLiteral $ choice [ 
                  oneOf "fF" *> return FloatLiteral
                , oneOf "dD" *> return DoubleLiteral
                ]
        let suffixLength = length suffix
        let prefixPower  = 10 ^ suffixLength :: Integer
        let iprefix = read prefix :: Integer
        let isuffix = if (suffixLength == 0) then 0 else read suffix :: Integer
        let exponentPower = (if exponent > 0 then 10 ^ exponent else 1) % (if exponent < 0 then 10^(-exponent) else 1)
        let rational = ((iprefix * prefixPower + isuffix) % prefixPower) * exponentPower
        return (FloatingLiteral rational literalType)


parseBooleanLiteral :: Parser LiteralExpression
parseBooleanLiteral = choice 
        [ m_reserved "true"  *> return (BooleanLiteral True)
        , m_reserved "false" *> return (BooleanLiteral False)
        ]

parseCharLiteral :: Parser LiteralExpression
parseCharLiteral = (char '\'' *> choice [
          parseEscapedChar
        , satisfy (not . flip elem "'\\")
        ] <* char '\'') >>= return . CharacterLiteral

-- TODO: Support double line terminator
parseEscapedChar :: Parser Char
parseEscapedChar = char '\\' *> choice [
          char 'b' *> return '\b'
        , char 's' *> return ' '
        , char 't' *> return '\t'
        , char 'n' *> return '\n'
        , char 'f' *> return '\f'
        , char 'r' *> return '\r'
        , char '"' *> return '"'
        , char '\'' *> return '\''
        , char '\\' *> return '\\'
        , parseOctalEscapedChar
        ]

parseOctalEscapedChar :: Parser Char
parseOctalEscapedChar = choice [
          try $ tripleOctalEscapedChar
        , try $ doubleOctalEscapedChar
        ,       singleOctalEscapedChar
        ]

singleOctalEscapedChar :: Parser Char
singleOctalEscapedChar = do
        char1 <- oneOf "01234567"
        let num1 = read [char1] :: Int
        return . Char.chr $ num1

doubleOctalEscapedChar :: Parser Char
doubleOctalEscapedChar = do
        char1 <- oneOf "01234567"
        char2 <- oneOf "01234567"
        let num1 = read [char1] :: Int
        let num2 = read [char2] :: Int
        let num = num2 + num1 * 8
        return . Char.chr $ num

tripleOctalEscapedChar :: Parser Char
tripleOctalEscapedChar = do
        char1 <- oneOf "0123"
        char2 <- oneOf "01234567"
        char3 <- oneOf "01234567"
        let num1 = read [char1] :: Int
        let num2 = read [char2] :: Int
        let num3 = read [char3] :: Int
        let num = num3 + num2 * 8 + num1 * 64
        return . Char.chr $ num

parseStringLiteral :: Parser LiteralExpression
parseStringLiteral = do
        _ <- char '"'
        content <- manyTill (choice
                [ parseEscapedChar
                , satisfy (not . flip elem "\r\n\"\\")
                ]) (char '"')
        return (StringLiteral content)

parseTextLiteral :: Parser LiteralExpression
parseTextLiteral = do
        _ <- string' "\"\"\""
        _ <- space `manyTill` endOfLine
        content <- (choice [ parseEscapedChar , satisfy (not . flip elem "\\") ] ) `manyTill` (string' "\"\"\"")
        return . TextLiteral $ content
